(module Math mzscheme
  (require (lib "Syntax.ss" "Mead"))
  (provide pi phi golden e fuzz
	         dotProduct crossProduct normalize vectorLength vectorScale
                 deg2rad rad2deg sqr sqrDifference distance difference sum vectorSum vectorDifference dirXform
		 area perimeter
                 2to3d 3to2d triangulate triangulateUV last closeList polygon semigon openList polyXform
		 blend blend0 morph splitBezier
		 ;area pts2edges edges2pts polyNormal triList poly2tri orientCCW
                 )
(define pi (acos -1))
(define phi (/ (+ 1 (sqrt 5)) 2))
(define golden (/ phi))
(define e (exp 1))
(define fuzz 0.000001)

;; dot product of two vectors of similar dimension
(define (dotProduct v0 v1)
  (apply + (map (lambda (a b) (* a b)) v0 v1)))

(define (vectorLength v)
  (sqrt (dotProduct v v)))

(define (vectorScale v s)
  (cond ([number? v] (* s v))
        ([list? v] (map (lambda (x) (vectorScale x s)) v))))

(define (crossProduct v0 v1)
  (let ([x0 (car v0)]
        [y0 (cadr v0)] 
        [z0 (caddr v0)]
        [x1 (car v1)]
        [y1 (cadr v1)]
        [z1 (caddr v1)])
    (list (- (* y0 z1) (* z0 y1)) (- (* z0 x1) (* x0 z1)) (- (* x0 y1) (* y0 x1)))))
 
(define (normalize v)
  (let ([length (sqrt (+ (sqr (car v)) (sqr (cadr v)) (sqr (caddr v))))])
    (if (< length fuzz) '(0 1 0)
        (list (/ (car v) length) (/ (cadr v) length) (/ (caddr v) length)))))

(define (deg2rad a)
  (/ (* pi a) 180))
(define (rad2deg a)
  (/ (* 180 a) pi))

(define (sqr a)
  (* a a))

(define (sqrDifference a b)
  (sqr (- a b)))

(define (distance a b)
  (cond
    ([number? a] (sqrt (sqrDifference a b)))
    ([list? a] (sqrt (apply + (map sqrDifference a b))))))

(define (difference a b)
  (cond
    ([number? a] (- a b))
    ([list? a] (map difference a b))))

(define (sum a b)
  (cond ([number? a] (+ a b))
        ([list? a] (map sum a b))))

(define (vectorSum a b)
  (sum a b))
  
(define (vectorDifference a b)
  (difference a b))
  

(define (dirXform v t)
  (difference (tell1 t mapPoint v) (tell1 t mapPoint '(0 0 0))))

(define (last l)
  (cond ([null? l] '())
        ([null? (cdr l)] (car l))
        (else (last (cdr l)))
        )
  )

(define (2to3d l)
  (if (null? l) '() (cons (append (car l) '(0)) (2to3d (cdr l)))))

(define (closeList p)
  (if (null? p) p
      (let* ([a (car p)]
             [l (list-ref p (- (length p) 1))])
        (if (< (distance a l) fuzz) p
            (append p (list a))))))
  

(define (perimeter0 start l)
  (if (null? l) 0
      (+ (distance start (car l))
	 (perimeter0 (car l) (cdr l)))))

(define (perimeter l)
  (if (null? l) 0
      (perimeter0 (last l) l)))

; Computing the area of a 2D polygon is based on computing the area under a line segment.
; The value is positive if we traverse the polygon in counterclockwise order, negative otherwise.
; Recursive helper.
(define (area0 p0 p1 rest)
  (let* ([a0 (* 0.5 (- (* (car p0) (cadr p1)) (* (car p1) (cadr p0))))])
    (if (null? rest) a0
        (+ a0 (area0 p1 (car rest) (cdr rest))))))

; Computes the area of a 2D polygon
(define (area p)
  (let* ([d (length (car p))]
	 [pg (closeList p)])
    (if (= 2 d)
	(area0 (car pg) (cadr pg) (cddr pg))))
)

; Returns a polygon in CCW order - canonical for our LH universe
(define (orientCCW p)
  (if (> (area p) 0) p (reverse p))
  )

; a is not a right turn if the area of a-b-c is non-negative
(define (convex a b c)
  (>= (area (list a b c)) 0))

; b is a left turn if the area of a-b-c is positive.
(define (concave a b c) (not (convex a b c)))

; (inefficient) triangle containment.  does not include boundary.
(define (triContains a b c p)
  (and (concave a p b)
       (concave b p c)
       (concave c p a)))

; Compute a list of concave polygon vertices.
; recursive helper function
(define (concaves0 a b c rest)
  (if (concave a b c)
      (if (null? rest) (list b)
          (cons b (concaves0 b c (car rest) (cdr rest))))
      (if (null? rest) '()
          (concaves0 b c (car rest) (cdr rest)))))

; computing the list of concave vertices
;; assumes that l is a full-list: closed, with repeated second point
(define (concaves l)
  (concaves0 (car l) (cadr l) (caddr l) (cdddr l)))

; returns #t if any of the vertices are within a-b-c
(define (triContainsAny a b c cvs)
  (if (null? cvs) #f
      (or (triContains a b c (car cvs)) (triContainsAny a b c (cdr cvs)))))

; finds an ear - a triangle that may be lopped off from a polygon
; Theorem: there is always at least one of these
(define (findEar i a b c rest cvs)
  (if (and (convex a b c) (not (triContainsAny a b c cvs))) i
      (if (null? rest) #f
          (findEar (+ 1 i) b c (car rest) (cdr rest) cvs))))

(define (ear p)
  (let* ([fl (append (closeList p) (list (cadr p)))]
         [n (- (length fl) 2)]
         [cvs (concaves fl)]
         [a (car fl)]
         [b (cadr fl)]
         [c (caddr fl)]
         [rest (cdddr fl)])
    (remainder (findEar 1 a b c rest cvs) n)))

; convert a polygon (a list of points) into a list of edge vectors
(define (edges0 a b rest)
  (if (null? rest) (list (difference b a))
      (cons (difference b a) (edges0 b (car rest) (cdr rest)))))

(define (pts2edges p)
  (let* ([pl (closeList p)])
    (edges0 (car pl) (cadr pl) (cddr pl))
    )
  )

(define (edges2pts0 psum l)
  (if (null? l)
      '()
      (let ([newPartial (sum psum (car l))])
	(cons newPartial (edges2pts0 newPartial (cdr l))))))

(define (edges2pts l)
  (cons '(0 0) (edges2pts0 '(0 0) l)))

; 
(define (polyNormal el)
  (do ([n (crossProduct (car el) (cadr el)) (crossProduct (car el) (cadr el))]
       [el (cdr el) (cdr el)]
       [e0 (car el) (car el)])
    ((> (vectorLength n) 0)
     (let* ([u (normalize e0)]
	    [nn (normalize n)]
	    [v (normalize (crossProduct nn u))])
       (list nn u v)))))

(define (3to2d p)
  (let* ([el (pts2edges p)]
         [newBasis (polyNormal el)]
         [u (cadr newBasis)]
         [v (caddr newBasis)]
         [newEl (map (lambda (e)
                       (list (dotProduct u e) (dotProduct v e))) el)]
         )
    (let ([p (edges2pts newEl)])
      (if (> (area p) 0) p
	  (map (lambda (x)
		 (list (car x) (- (cadr x)))) p)
	  ))))

(define (list-rem p n)
  (if (= n 0)
      (cdr p)
      (cons (car p) (list-rem (cdr p) (- n 1)))))

(define (poly2tri p)
  (let* ([n (length p)])
    (if (= n 3)
        '(0)
        (let* ([target (ear p)]
               [newP (list-rem p target)])
          (cons target (poly2tri newP))))))

(define (triList n p order)
  (if (= n 3) (list p)
      (let* ([loc (car order)]
             [before (remainder (+ loc n -1) n)]
             [after (remainder (+ loc 1) n)]
             [tri (list (list-ref p before) (list-ref p loc) (list-ref p after))]
             [newP (list-rem p loc)]
             )
        (cons tri (triList (- n 1) newP (cdr order))))))

(define (triangulateUV p uv)
  (let* ([p0 (closeList (openList p))]
	 [uv0 (closeList (openList uv))]
	 [p2 (3to2d p0)]
         [order (poly2tri p2)]
         [n (- (length p0) 1)]
         )
    (list (triList n p0 order)
	  (triList n uv0 order)))
)  
(define (triangulate p)
  (let* ([p0 (closeList (openList p))]
	 [p1 (3to2d p0)]
         [order (poly2tri p1)]
         [n (- (length p0) 1)]
         )
        (triList n p0 order)
))

; 
(define (polygon n . radius)
  (let ([da (deg2rad (/ 360 n))]
	[r (if (null? radius) 50 (car radius))])
    (do ([m 0 (+ 1 m)]
	 [l '() (append l (list (list (* r (cos (* da m)))
				      (* r (sin (* da m))))))]
	 )
	 ((= m n) l)
      )
    )
  )
(define (semigon n . radius)
  (let ([da (deg2rad (/ 180 n))])
    (do ([r (if (null? radius) 50 (car radius)) r]
	 [m 0 (+ 1 m)]
	 [l '() (append l (list (list (* r (sin (* da m)))
				      (* r (- (cos (* da m))))
				      )
				)
			)
	    ]
	 )
	((> m n) l)
      )
    )
  )

(define (openList0 head first rest)
  (cond ([null? rest] (if (> fuzz (distance first head)) '()
			  (list first)))
	(else (let* ([next (car rest)])
		(if (> fuzz (distance first next))
		    (openList0 head next (cdr rest))
		    (cons first (openList0 head next (cdr rest))))))))

(define (openList l)
  (cond ([> 2 (length l)] l)
	(else (openList0 (car l) (car l) (cdr l)))))

(define (polyXform l t)
  (if (null? l)
      '()
      (cons (tell1 t mapPoint (car l)) (polyXform (cdr l) t))))

(define (rand low high)
  (+ low (* random (- high low)))
)

(define (blend0 coeffs args)
  (let ([first (car args)])
    (cond ([null? first] '())
	  ([number? first] (apply + (map * coeffs args)))
	  ([list? first] (cons (blend0 coeffs (map car args))
			       (blend0 coeffs (map cdr args)))))
    )
  )

(define (blend t . args)
  (let* ([1mt (- 1 t)]
	 [l (length args)])
    (cond ([= l 1] (car args))
	  ([= l 2] (blend0 (list 1mt t) args))
	  ([= l 3] (blend0 (list (* 1mt 1mt) (* 2 1mt t) (* t t)) args))
	  ([= l 4] (blend0 (list (* 1mt 1mt 1mt) (* 3 1mt 1mt t)
				 (* 3 1mt t t) (* t t t)) args)))))

(define (morph0 n a b)
  (if (<= n 0)
      (list b)
      (cons a
	    (morph0 (- n 1) (blend (/ 1.0 n) a b) b))
      )
  )


(define (morph n . rest)
  (let* ([pts (morph0 n 0 1)])
    (map (lambda (pt)
	   (apply blend (cons pt rest))) pts))
  )

(define (splitBezier pct a b c d)
  (let* ([s0 (blend pct a b)]
         [m  (blend pct b c)]
         [t1 (blend pct c d)]
         [s1 (blend pct s0 m)]
         [t0 (blend pct m t1)]
         [e  (blend pct s1 t0)])
    (list (list a s0 s1 e)
          (list e t0 t1 d))))
)
