(module Meshes mzscheme
  (require (lib "Syntax.ss" "Mead")
           (lib "Math.ss" "Mead")
           (lib "Object.ss" "Mead")) ;; this incorporates Materials, as well.
  (provide Mesh PatchMesh
	   HeightField extrude sweep loft)

(define (ptEqv? a b)
  (< (distance a b) fuzz))

(define (ptLookup p l)
  (lookup p l ptEqv?))

(define (addPt p l)
  (let* ((i [ptLookup p l]))
    (if i (cons i l)
        (begin
          (cons (length l) (append l (list p)))))))

(define (addUv p l)
  (let* ((i [uvLookup p l]))
    (if i (cons i l)
        (begin
          (cons (length l) (append l (list p)))))))

(define (lookup p l f)
  (lookup2 p l f 0))

(define (lookup2 p l f n)
  (cond ([null? l] #f)
        ([f p (car l)] n)
        (else (lookup2 p (cdr l) f (+ 1 n)))))

(define (uvEqv? a b)
  (< (distance a b) fuzz))
(define (uvLookup p l)
  (lookup p l uvEqv?))

(define Mesh
  (class Primitive
    (inherit selected attr getAttr material getMaterial getXform
             scale translate xRot yRot zRot
             xyMirror xzMirror yzMirror)
    (public register)
    (define (register v)
      (let* ([a (getAttr 'vertices)]
             [p (addPt v a)])
        (attr 'vertices (cdr p))
        (car p)))
    
    (define/public (uvRegister v)
      (let* ([a (getAttr 'uvVertices)]
             [uv (addUv v a)])
        (attr 'uvVertices (cdr uv))
        (car uv)))
    
    (define/public (addTri t)
      (let* ([v0 (register (car t))]
             [v1 (register (cadr t))]
             [v2 (register (caddr t))])
        (attr 'triangles (cons (list v0 v1 v2) (getAttr 'triangles)))))
    
    (define/public (addUVTri t uv)
      (let* ([v0 (register (car t))]
             [v1 (register (cadr t))]
             [v2 (register (caddr t))]
             [uv0 (uvRegister (car uv))]
             [uv1 (uvRegister (cadr uv))]
             [uv2 (uvRegister (caddr uv))])
        (attr 'triangles (cons (list v0 v1 v2) (getAttr 'triangles)))
        (attr 'uvTriangles (cons (list uv0 uv1 uv2) (getAttr 'uvTriangles)))))

    (define/public (addPoly p)
      (let* ([op (openList p)]
	     [l (length op)])
	(if (= l 3)
	    (addTri op)
	    (if (> l 3)
		(let ([tL (triangulate op)])
		  (map (lambda (t) (addTri t)) tL))
		)
	    )
	)
      )

    (define/public (addUVPoly p uv)
      (let* ([triLists (triangulateUV p uv)]
            [tList (car triLists)]
            [uvList (cadr triLists)]
            )
        (map (lambda (t uv) (addUVTri t uv)) tList uvList))
      )
    (define/public (getVertices)
      (getAttr 'vertices))
    (define/public (getTriangles)
      (getAttr 'triangles))
    (define/public (getUVVertices)
      (getAttr 'uvVertices))
    (define/public (getUVTriangles)
      (getAttr 'uvTriangles))

    (define/public (registerVertices l)
      (attr 'vertices l))
    (define/public (registerTriangles l)
      (attr 'triangles l))
    (define/public (registerUVVertices l)
      (attr 'uvVertices l))
    (define/public (registerUVTriangles l)
      (attr 'uvTriangles l))
    
    (define/override (POVDescribe context)
      (if (selected context)
          (let* ([vl (getVertices)]
                 [uvl (getUVVertices)]
                 [nv (length vl)]
                 [tl (getTriangles)]
                 [nt (length tl)]
                 [uvtl (getUVTriangles)]
                 [t (getXform)]
                 [m (getMaterial)])
            (printf "mesh2 {\n")
            (printf "vertex_vectors { ~a" nv)
            (do ([vl vl (cdr vl)])
              ([null? vl])
              (printf ",")
              (printPoint (car vl)))
            (printf "}\n")
            (if (not (null? uvl)) 
                (begin
                  (printf "uv_vectors { ~a" (length uvl))
                  (do ([uvl uvl (cdr uvl)])
                    ([null? uvl])
                    (printf ",")
                    (printPoint (car uvl)))
                  (printf "}\n")))
            (printf "face_indices { ~a" nt)
            (do ([tl tl (cdr tl)])
              ([null? tl])
              (printf ",")
              (printPoint (car tl)))
            (printf "}\n")
          
            (if (not (null? uvtl))
                (begin
                  (printf "uv_indices { ~a" (length uvtl))
                  (do ([uvtl uvtl (cdr uvtl)])
                    ([null? uvtl]) (printf ",")
                    (printPoint (car uvtl)))
                  (printf "}\n")))
            (printf "inside_vector <0,0,1>\n")

            (super POVDescribe context)
            (printf "}\n")
            )
          )
      )
    
    (begin
      (super-new)
      (registerVertices '())
      (registerTriangles '())
      (registerUVVertices '())
      (registerUVTriangles '())
      )))

(define HeightField
  (class Primitive
    (inherit attr getAttr selected description scale translate)
    (public image getImage imageType getImageType
	    smooth smooth? clipLevel getClipLevel)
    (define (image i)
      (attr 'image i)
      )
    (define (getImage)
      (getAttr 'image)
      )
    (define (imageType t)
      (attr 'imageType t)
      )
    (define (getImageType)
      (getAttr 'imageType)
      )
    (define (smooth b)
      (attr 'smooth b)
      )
    (define (smooth?)
      (getAttr 'smooth)
      )
    (define (clipLevel v)
      (attr 'clipLevel v)
      )
    (define (getClipLevel)
      (getAttr 'clipLevel)
      )
    (define/override (POVDescribe context)
      (if (selected context)
	  (let* ([s (smooth?)]
		 [cl (getClipLevel)]
		 )
	    (printf "height_field { ~a \"~a\" " (getImageType) (tell environment (makeImagePath (getImage))))
	    (if (smooth?) (printf " smooth"))
	    (if cl (printf " water_level ~a" cl))
	    (printf "\n")
	    (super POVDescribe context)
	    (printf "}\n")
	    )
	  )
      )
    (begin
      (super-new)
      (description "An image-based height field.")
      (scale 100 100 100)
      (translate -50 -50 -50)
      (imageType 'png)
      )
    )
  )
	    
(define PatchMesh
  (class Primitive
    (inherit selected attr getAttr material getMaterial getXform
             scale translate xRot yRot zRot
             xyMirror xzMirror yzMirror)
    (public register uvRegister
	    registerVertices registerPatches
	    registerUVVertices registerUVPatches
	    getVertices getPatches
	    getUVVertices  getUVPatches 
	    addPatch addUVPatch
	    refinement getRefinement
	    POVDescribePatch POVDescribeUVPatch)
    
    (define (register v)
      (let* ([a (getAttr 'vertices)]
             [p (addPt v a)])
        (attr 'vertices (cdr p))
        (car p))
      )

    (define (uvRegister v)
      (let* ([a (getAttr 'uvVertices)]
             [uv (addUv v a)])
        (attr 'uvVertices (cdr uv))
        (car uv)
	)
      )

    (define (registerVertices vl)
      (attr 'vertices vl)
      )

    (define (registerPatches ps)
      (attr 'patches ps)
      )

    (define (registerUVVertices uvl)
      (attr 'uvVertices uvl)
      )

    (define (registerUVPatches uvps)
      (attr 'uvPatches uvps)
      )

    (define (getUVVertices)
      (getAttr 'uvVertices)
      )
    
    (define (getUVPatches)
      (getAttr 'uvPatches)
      )

    (define (getVertices)
      (getAttr 'vertices)
      )
    
    (define (getPatches)
      (getAttr 'patches)
      )
    
    (define (addPatch t)
      (let* ([v (list->vector t)])
        (do ([i 0 (+ 1 i)])
          ((>= i 16))
          (vector-set! v i (register (vector-ref v i))))
        (registerPatches (cons v (getPatches))))
      )

    (define (refinement level)
      (attr 'refinement level)
      )
    (define (getRefinement)
      (getAttr 'refinement)
      )

    (define (addUVPatch t uv)
      (let* ([v (list->vector t)]
	     [v0 (uvRegister (car uv))]
	     [v1 (uvRegister (cadr uv))]
	     [v2 (uvRegister (caddr uv))]
	     [v3 (uvRegister (car (cdddr uv)))]
	     )
	; register patch vertices & update polygon
        (do ([i 0 (+ 1 i)])
	    ((>= i 16))
          (vector-set! v i (register (vector-ref v i))))
        (registerPatches (cons v (getPatches)))
	(registerUVPatches (cons (list v0 v1 v2 v3) (getUVPatches)))
	)
      )

    (define (POVDescribePatch p context)
      (let* ([p (list-ref (getPatches) p)]
             [v (getVertices)]
             [t (getXform)]
             [m (getMaterial)]
	     [r (or (getRefinement) 3)])
        (printf "bicubic_patch { type 0 flatness 0.01 u_steps ~a v_steps ~a\n" r r)
        (do ([i 0 (+ 1 i)])
          ((>= i 16))
          (if (not (= i 0)) (printf ","))
          (printPoint (list-ref v (vector-ref p i)))
          )
        (printf "\n")
        ;(printf "inside_vector <0,0,1>\n")
        (super POVDescribe context)
        (printf "}\n")
        )
      )

    (define (POVDescribeUVPatch p context)
      (let* ([uvp (list-ref (getUVPatches) p)]
	     [p (list-ref (getPatches) p)]
             [v (getVertices)]
	     [uvv (getUVVertices)]
	     [v0 (list-ref uvv (car uvp))]
	     [v1 (list-ref uvv (cadr uvp))]
	     [v2 (list-ref uvv (caddr uvp))]
	     [v3 (list-ref uvv (car (cdddr uvp)))]
             [t (getXform)]
             [m (getMaterial)]
	     [r (or (getRefinement) 3)])
        (printf "bicubic_patch { type 0 flatness 0.01 u_steps ~a v_steps ~a\n" r r)
	(printf "uv_vectors ")
	(printPoint v0) (printf ",") 
	(printPoint v1) (printf ",")
	(printPoint v2) (printf ",")
	(printPoint v3) (printf "\n")
        (do ([i 0 (+ 1 i)])
          ((>= i 16))
          (if (not (= i 0)) (printf ","))
          (printPoint (list-ref v (vector-ref p i)))
          )
        (printf "\n")
        ;(printf "inside_vector <0,0,1>\n")
        (super POVDescribe context)
        (printf "}\n")
        )
      )
    
    (define/override (POVDescribe context)
      (if (selected context)
	  (if (not (null? (getUVPatches)))
	      (let* ([n (length (getPatches))])
		(do ([i 0 (+ 1 i)])
		    ([>= i n])
		  (POVDescribeUVPatch i context)))
	      (let* ([n (length (getPatches))])
		(do ([i 0 (+ 1 i)])
		    ([>= i n])
		  (POVDescribePatch i context)))
	      )
	  )
      )
    
    (begin
      (super-new)
      (refinement 3)
      (registerVertices '())
      (registerPatches '())
      (registerUVVertices '())
      (registerUVPatches '())
      )
    )
 )

;; Implementation of triMesh extrudes
(define (extrude1 head tail result)
  (if (or (< (length head) 2) (< (length tail) 2)) result
    (let* ([h0 (car head)]
           [h1 (cadr head)]
           [t1 (cadr tail)]
           [t0 (car tail)])
      (tell result (addPoly (list h0 h1 t1 h0)))
      (tell result (addPoly (list h0 t1 t0 h0)))
      (extrude1 (cdr head) (cdr tail) result))))

(define (extrude0 head xs result)
  (if (null? xs)
      (begin
        (tell result (addPoly head))
        result)
      (let* ([x (car xs)]
             [tail (polyXform head x)]
             )
        (extrude1 head tail result)
        (extrude0 tail (cdr xs) result))))

(define (extrude p . xs)
   (let* ((cap0 (closeList p))
          (result (new Mesh))
        )
    (tell result (addPoly cap0))
     (extrude0 cap0 xs result)))

(define (caplessExtrude mesh poly xf)
  (let ([poly2 (polyXform poly xf)])
    (do ([lastp (last poly) (car poly)]
         [lastp2 (last poly2) (car poly2)]
         [poly poly (cdr poly)]
         [poly2 poly2 (cdr poly2)]
         )
      ([null? poly] mesh)
      (let* ([p (car poly)]
             [p2 (car poly2)]
	     )
	(tell mesh
	      (addTri (list p p2 lastp2))
	      (addTri (list lastp2 lastp p))
	      )
        )
      )
    )
  )
      
(define (sweep0 mesh poly i n xf)
  (if (>= i n) mesh
      (begin
        (caplessExtrude mesh poly xf)
        (sweep0 mesh (polyXform poly xf) (+ i 1) n xf)
        )
      )
  )

(define (sweep shape count)
  (if (> count 2)
      (let* ([m (new Mesh)]
             [rotation (yRot (/ 360 count))]
             )
        (sweep0 m shape 0 count rotation)
        )
      )
  )

(define (loft0 head ps result)
  (if (null? ps)
      (begin
        (tell result (addPoly head))
        result)
      (let* ([tail (car ps)]
             )
        (extrude1 head tail result)
        (loft0 tail (cdr ps) result))))

(define (loft . ps)
   (let* ([p (car ps)]
	  [cap (closeList p)]
          [result (new Mesh)]
        )
    (tell result (addPoly cap))
     (loft0 cap (cdr ps) result)))
)
