;; Definitions of Objects, Materials, Transforms and other primitives.
;; (c) 2008 duane a. bailey
;; these utility functions allow the update of attribute lists.
;; they should probably be part of the MeadObject class, as private functions.
(module Object mzscheme
	(require (lib "Syntax.ss" "Mead")
		 (lib "Math.ss" "Mead")
		 )
	(provide update copy printPoint printXform
		 MeadObject Environment environment Transform Transformable Primitive Context Reference Group
		 describe translate scale xRot yRot zRot xyMirror xzMirror yzMirror
		 compose xform withoutTag withTag
		 hsv2rgb rgb2hsv
		 MeadColor Material emptyMaterial defaultMaterial)

	;; update a key's value in a list
	(define (update k v l)
	  (let ((kv (assoc k l)))
	    (if kv (begin (set-car! (member kv l) (cons k v)) l)
		(cons (cons k v) l))))

	; make a copy of a list
	(define (copy l)
	  (cond ([null? l] l)
		([null? (cdr l)] (cons (car l) null))
		(else (cons (car l) (copy (cdr l))))))

	; print out a POV version of a point
	(define (printPoint p)
	  (do ([prefix "<" ","]
	       [p p (cdr p)])
	      ((null? p) (printf ">"))
	    (printf "~a~a" prefix (car p))))

	; print out a POV version of a transform
	(define (printXform prefix t)
	  (printf prefix)
	  (printf "<")
	  (do ([i 0 (+ 1 i)]) ((>= i 15))
	    (if (not (eq? 3 (modulo i 4)))
		(begin
		  (if (> i 0) (printf ", "))
		  (printf "~a" (vector-ref t i)))))
	  (printf ">\n")
	  )
	;; Top of the Mead object hierarchy.

	(define MeadObject
	  (class object%
					; public methods that may be overridden
		 (public describe description makeCopy attributes getAttributes attr getAttr)
		 
					; private fields
		 (define attrs null)
		 
					; method implementation
		 (define (attributes l)
		   (set! attrs l))
		 (define (getAttributes)
		   attrs)
		 
		 (define (getAttr key)
		   (let ((kv (assv key attrs)))
		     (and kv (cdr kv))))
		 
		 (define (attr key value)
		   (set! attrs (update key value attrs)))
		 
		 (define (description d)
		   (attr 'description d))
		 
		 (define (describe)
		   attrs)
		 
		 (define (makeCopy)
		   (let ((c (make-object MeadObject)))
		     (tell c (setAttributes (copy attrs)))
		     c))
		 
					; superclass initialization
		 (begin
		   (super-new)
		   (description "A Mead object"))))

	(define Transform
	  (class MeadObject
		 (inherit description attr getAttr)
		 (public xRot yRot zRot scale translate
			 yzMirror xzMirror xyMirror
			 mapPoint mapDirection
			 getHistory getXform 
			 thenXformWith prefixWith suffixWith
			 identity?)
		 (private inner)
		 
		 (define (inner a b c)
		   (+ (* a b) c))
		 
		 (define (identity?)
		   #f)
		 
		 (define (appendHistory t)
		   (attr 'history (append (getHistory) (list t))))
		 
		 (define (mapPoint p)
		   (let* ([x (car p)] 
			  [y (cadr p)]
			  [z (caddr p)]
			  [t (getAttr 'xform)]
			  [omega (+ (* (vector-ref t 3) x)
				    (* (vector-ref t 7) y)
				    (* (vector-ref t 11) z)
				    (vector-ref t 15))]
			  [rx (/ (+ (* (vector-ref t 0) x)
				    (* (vector-ref t 4) y)
				    (* (vector-ref t 8) z)
				    (vector-ref t 12)) omega)]
			  [ry (/ (+ (* (vector-ref t 1) x)
				    (* (vector-ref t 5) y)
				    (* (vector-ref t 9) z)
				    (vector-ref t 13)) omega)]
			  [rz (/ (+ (* (vector-ref t 2) x)
				    (* (vector-ref t 6) y)
				    (* (vector-ref t 10) z)
				    (vector-ref t 14)) omega)])
		     (list rx ry rz)))
		 (define (mapDirection v)
		   (difference (mapPoint v) (mapPoint '(0 0 0))))
		 
		 (define (yzMirror)
		   (appendHistory '(yzMirror))
		   (let* ([t (getAttr 'xform)])
		     (vector-set! t 0 (- (vector-ref t 0)))
		     (vector-set! t 4 (- (vector-ref t 4)))
		     (vector-set! t 8 (- (vector-ref t 8)))
		     (vector-set! t 12 (- (vector-ref t 12)))
		     )
		   )
		 
		 (define (xzMirror)
		   (appendHistory '(xzMirror))
		   (let* ([t (getAttr 'xform)])
		     (vector-set! t 1 (- (vector-ref t 1)))
		     (vector-set! t 5 (- (vector-ref t 5)))
		     (vector-set! t 9 (- (vector-ref t 9)))
		     (vector-set! t 13 (- (vector-ref t 13)))
		     )
		   )
		 
		 (define (xyMirror)
		   (appendHistory '(xyMirror))
		   (let* ([t (getAttr 'xform)])
		     (vector-set! t 2 (- (vector-ref t 2)))
		     (vector-set! t 6 (- (vector-ref t 6)))
		     (vector-set! t 10 (- (vector-ref t 10)))
		     (vector-set! t 14 (- (vector-ref t 14)))
		     )
		   )
		 
		 (define (xRot angle)
		   (appendHistory (list 'xRot angle))
		   (let* ([a (deg2rad angle)]
			  [c (cos a)]
			  [s (sin a)]
			  [t (getAttr 'xform)]
			  [t1 (vector-ref t 1)]
			  [t5 (vector-ref t 5)]
			  [t9 (vector-ref t 9)]
			  [t13 (vector-ref t 13)])
		     (vector-set! t 1 (- (* c t1) (* s (vector-ref t 2))))
		     (vector-set! t 2 (+ (* s t1) (* c (vector-ref t 2))))
		     (vector-set! t 5 (- (* c t5) (* s (vector-ref t 6))))
		     (vector-set! t 6 (+ (* s t5) (* c (vector-ref t 6))))
		     (vector-set! t 9 (- (* c t9) (* s (vector-ref t 10))))
		     (vector-set! t 10 (+ (* s t9) (* c (vector-ref t 10))))
		     (vector-set! t 13 (- (* c t13) (* s (vector-ref t 14))))
		     (vector-set! t 14 (+ (* s t13) (* c (vector-ref t 14))))
		     )
		   )
		 
		 (define (yRot angle)
		   (appendHistory (list 'yRot angle))
		   (let* ([a (deg2rad angle)]
			  [c (cos a)]
			  [s (sin a)]
			  [t (getAttr 'xform)]
			  [t0 (vector-ref t 0)]
			  [t4 (vector-ref t 4)]
			  [t8 (vector-ref t 8)]
			  [t12 (vector-ref t 12)])
		     (vector-set! t 0 (+ (* c t0) (* s (vector-ref t 2))))
		     (vector-set! t 2 (- (* c (vector-ref t 2)) (* s t0)))
		     (vector-set! t 4 (+ (* c t4) (* s (vector-ref t 6))))
		     (vector-set! t 6 (- (* c (vector-ref t 6)) (* s t4)))
		     (vector-set! t 8 (+ (* c t8) (* s (vector-ref t 10))))
		     (vector-set! t 10 (- (* c (vector-ref t 10)) (* s t8)))
		     (vector-set! t 12 (+ (* c t12) (* s (vector-ref t 14))))
		     (vector-set! t 14 (- (* c (vector-ref t 14)) (* s t12)))
		     ))
		 
		 (define (zRot angle)
		   (appendHistory (list 'zRot angle))
		   (let* ([a (deg2rad angle)]
			  [c (cos a)]
			  [s (sin a)]
			  [t (getAttr 'xform)]
			  [t0 (vector-ref t 0)]
			  [t4 (vector-ref t 4)]
			  [t8 (vector-ref t 8)]
			  [t12 (vector-ref t 12)])
		     (vector-set! t 0 (- (* c t0) (* s (vector-ref t 1))))
		     (vector-set! t 1 (+ (* s t0) (* c (vector-ref t 1))))
		     (vector-set! t 4 (- (* c t4) (* s (vector-ref t 5))))
		     (vector-set! t 5 (+ (* s t4) (* c (vector-ref t 5))))
		     (vector-set! t 8 (- (* c t8) (* s (vector-ref t 9))))
		     (vector-set! t 9 (+ (* s t8) (* c (vector-ref t 9))))
		     (vector-set! t 12 (- (* c t12) (* s (vector-ref t 13))))
		     (vector-set! t 13 (+ (* s t12) (* c (vector-ref t 13))))
		     ))
		 
		 (define (scale x y z)
		   (if (or (= x 0) (= y 0) (= z 0))
		       (raise-user-error 'scale 
			 "A zero value was found in scale; probably should be 1"))
		   (appendHistory (list 'scale x y z))
		   (let ([t (getAttr 'xform)])
		     (vector-set! t 0 (* (vector-ref t 0) x))
		     (vector-set! t 1 (* (vector-ref t 1) y))
		     (vector-set! t 2 (* (vector-ref t 2) z))
		     (vector-set! t 4 (* (vector-ref t 4) x))
		     (vector-set! t 5 (* (vector-ref t 5) y))
		     (vector-set! t 6 (* (vector-ref t 6) z))
		     (vector-set! t 8 (* (vector-ref t 8) x))
		     (vector-set! t 9 (* (vector-ref t 9) y))
		     (vector-set! t 10 (* (vector-ref t 10) z))
		     (vector-set! t 12 (* (vector-ref t 12) x))
		     (vector-set! t 13 (* (vector-ref t 13) y))
		     (vector-set! t 14 (* (vector-ref t 14) z))
		     )
		   )
		 
		 (define (translate x y z)
		   (appendHistory (list 'translate x y z))
		   (let* ([t (getAttr 'xform)]
			  [v3 (vector-ref t 3)]
			  [v7 (vector-ref t 7)]
			  [v11 (vector-ref t 11)]
			  [v15 (vector-ref t 15)])
		     (vector-set! t 0 (inner v3 x (vector-ref t 0)))
		     (vector-set! t 1 (inner v3 y (vector-ref t 1)))
		     (vector-set! t 2 (inner v3 z (vector-ref t 2)))
		     
		     (vector-set! t 4 (inner v7 x (vector-ref t 4)))
		     (vector-set! t 5 (inner v7 y (vector-ref t 5)))
		     (vector-set! t 6 (inner v7 z (vector-ref t 6)))
		     
		     (vector-set! t 8 (inner v11 x (vector-ref t 8)))
		     (vector-set! t 9 (inner v11 y (vector-ref t 9)))
		     (vector-set! t 10 (inner v11 z (vector-ref t 10)))
		     
		     (vector-set! t 12 (inner v15 x (vector-ref t 12)))
		     (vector-set! t 13 (inner v15 y (vector-ref t 13)))
		     (vector-set! t 14 (inner v15 z (vector-ref t 14)))
		     )
		   )
		 
		 (define (getHistory)
		   (getAttr 'history))
		 (define (getXform)
		   (getAttr 'xform))
		 (define (prefixWith other)
		   (let* ([oh (tell other (getHistory))]
			  [h  (getHistory)]
			  [newXform (tell other (thenXformWith this))])
		     (attr 'xform newXform)
		     (attr 'history (append oh h))))
		 
		 (define (suffixWith other)
		   (let* ([oh (tell1 other getHistory)]
			  [h (getHistory)]
			  [newXform (thenXformWith other)])
		     (attr 'xform (tell1 newXform getXform))
		     (attr 'history (append h oh))))
		 
		 (define (thenXformWith other)
		   ;; compose this transform with another
		   (let* ([ot (tell other (getXform))]
			  [t (getAttr 'xform)]
			  [newXform (new Transform)]
			  [nt (tell newXform (getAttr 'xform))]
			  )
		     (do ([i 0 (+ 1 i)])
			 ((>= i 16) nt)
		       (do ([j (modulo i 4) (+ j 4)]
			    [k (- i (modulo i 4)) (+ k 1)]
			    [sum 0 sum])
			   ((>= j 16) (vector-set! nt i sum))
			 (set! sum (+ (* (vector-ref ot j) (vector-ref t k)) sum))))
		     newXform))
		 
		 (define/public (transform t)
		   (attr 'xform t))
		 
		 (define/public (normalize)
		   (let* ([t (getAttr 'xform)]
			  [omega (vector-ref 'xform 15)])
		     (if (not (eqv? omega 1))
			 (do ([i 0 (+ 1 i)])
			     ((>= i 15))
			   (vector-set! t i (/ (vector-ref t i) omega))))))
		 
		 (define/public (POVDescribe . context)
					;(tell1 context pushTransform this)
					;(let* ([t (tell1 context total)])
					;  (printXform "matrix" (tell1 t getXform)))
					; (tell1 context popTransform))
		   (printXform "matrix" (getXform)))
		 
		 (begin
		   (super-new)
		   (description "A Mead transform")
		   (attr 'history null)
		   (attr 'xform (vector 1 0 0 0
					0 1 0 0
					0 0 1 0
					0 0 0 1)))
		 )
	  )

; The environment is responsible for establishing a connection with the
; underlying operating system. The default Envrionment, environment, one
; is able to describe where surface and grain images are stored, where
; the resulting project elements are stored, and where the POV settings
; files can be loaded from (library...)
;
; This is made complex by the fact that the model may not have been stored
; to disk at the time that it is constructed.  Thus, by default, the images
; are stored in a temporary directory, etc.
;
; More information about this interaction can be found in the mead-pov and
; mead-buildMovie scripts.
	(define Environment
	  (class MeadObject
		 (inherit attr getAttr description)
		 (public imageFolder getImageFolder makeImagePath
			 projectFolder getProjectFolder
			 libraryFolder getLibraryPath
			 POVDescribe POVProof)
		 (define (libraryFolder f)
		   (let ([p (getLibraryPath)])
		     (if p
			 (attr 'libraryPath (cons f p))
			 (attr 'libraryPath (list f))
			 )))
		 (define (getLibraryPath)
		   (getAttr 'libraryPath)
		   )
		 (define (imageFolder f)
		   (attr 'imageFolder f))
		 (define (getImageFolder)
		   (getAttr 'imageFolder)
		   )
		 (define (makeImagePath f)
		   (let* ([home (or (getenv "HOME") "/tmp")]
			  [ef (getImageFolder)]
			  [folder (or (and (string? ef) ef) "Images")]
			  [file (or (and (string? f) (< 0 (string-length f)) f) "untitled.png")])
		     (if (and (< 0 (string-length file)) (eq? #\/ (string-ref file 0)))
			 (begin
			   (set! home "")
			   (set! folder "")))
		     (if (and (< 0 (string-length folder)) (eq? #\/ (string-ref folder 0)))
			 (begin
			   (set! home "")))
		     (string-append home "/" folder "/" file)
		     )
		   )
		 (define (projectFolder f)
		   (attr 'projectFolder f))
		 (define (getProjectFolder)
		   (getAttr 'projectFolder)
		   )

		 (define (POVProof . args)
		   (POVDescribe)
		   )
		 
		 (define (POVDescribe)
					; written as part of scene initialization file
		   (if (getLibraryPath)
		       (begin
			 (do ([l (reverse (getLibraryPath)) (cdr l)])
			     ((null? l) #f)
			   (printf "Library_Path=\"~a\"\n" (car l))
			   )
			 )
		       )
		   )
		 (begin
		   (super-new)
		   (tell this 
			 (imageFolder "Images/")
			 (projectFolder "/tmp/")
			 )
		   )
		 )
	  )

	(object environment Environment)

	(define Transformable
	  (class MeadObject
		 (inherit attr getAttr)
		 (public tags getTags tag untag hasTag? selected
			 getXform centroid getCentroid
			 xRot yRot zRot scale translate
			 xyMirror xzMirror yzMirror
			 xform relativeXform absoluteXform
			 POVDescribe
			 )
		 (define (getTags)
			   (getAttr 'tags))
			 (define (tags tl)
			   (attr 'tags tl))
			 (define (tag . l)
			   (for-each (lambda (t)
				       (let ([tl (getTags)])
					 (if (not (member t tl)) (tags (cons t tl)))
					 )) l))
			 (define (untag . l)
			   (for-each (lambda (t)
				       (tags
					(let loop ([tl (getTags)])
					  (if (null? tl) tl
					      (if (equal? (car tl) t) (cdr tl)
						  (cons (car tl) (loop (cdr tl)))))))) l))
			 
			 (define (hasTag? t)
			   (and (member t (getTags)) #t))
			 
			 (define (selected context)
			   (let ([f (tell1 context getSelectionPredicate)])
			     (or (not f) (f this))))
			 
			 (define (xform t) (attr 'xform t))
			 (define (absoluteXform t) (attr 'xform t))
			 (define (relativeXform t) (tell (getXform) (suffixWith t)))
			 (define (getXform) (getAttr 'xform))
			 (define (centroid p) (attr 'centroid p))
			 (define (getCentroid) (getAttr 'centroid))
			 (define (xRot angle) (tell (getXform) (xRot angle)))
			 (define (yRot angle) (tell (getXform) (yRot angle)))
			 (define (zRot angle) (tell (getXform) (zRot angle)))
			 (define (scale x y z) (tell (getXform) (scale x y z)))
			 (define (translate x y z) (tell (getXform) (translate x y z)))
			 (define (xyMirror) (tell (getXform) (xyMirror)))
			 (define (xzMirror) (tell (getXform) (xzMirror)))
			 (define (yzMirror) (tell (getXform) (yzMirror)))
			 
			 (define (POVDescribe context)
			   (tell context (pushTransform (getXform)))
			   (let ([t (tell context (total))])
			     (tell t (POVDescribe context))
					;(printf " // centroid at ~a\n" (tell1 t mapPoint (getCentroid)))
			     )
			   (tell context (popTransform))
			   (tell context (pushTags (getTags)))
			   (let ([g (tell context (mergeTags))])
			     (if (not (null? g)) (printf " // tags: ~a\n" g)))
			   (tell context (popTags))
			   )
			 
			 (begin
			   (super-new)
			   (attr 'xform (new Transform))
			   (centroid '(0 0 0))
			   (tags null)
			   )
			 )
		 )

	  (define Primitive
	    (class Transformable
		   (inherit attr getAttr)
		   (define/public (material m)
		     (attr 'material m))
		   
		   (define/public (getMaterial)
		     (getAttr 'material))
		   
		   
		   (define/override (POVDescribe context)
		     ;; This section needs to be modified to support
		     ;; different transformations for materials and objects.
		     (tell context (pushMaterial (getMaterial)))
		     (tell (tell context (mergeMaterials)) (POVDescribe context))
		     (tell context (popMaterial))
		     (super POVDescribe context)
		     )
		   
		   (begin
		     (super-new)
		     (material emptyMaterial)
		     )))


	  (define Context
	    (class MeadObject
		   (inherit attr getAttr)
		   (define d 0)
		   (field (tStack null))
		   (field (mStack null))
		   (field (gStack null))
		   
		   (define/public (maxDepth d) (attr 'maxDepth d))
		   (define/public (getMaxDepth) (getAttr 'maxDepth))
		   (define/public (down) (set! d (+ 1 d)))
		   (define/public (up) (set! d (- 1 d)))
		   (define/public (getDepth) d)
		   (define/public (top)
		     (caar tStack))
		   (define/public (total)
		     (cadar tStack))
		   (define/public (topMat)
		     (caar mStack))
		   (define/public (mergeMaterials)
		     (cadar mStack))
		   (define/public (topTags)
		     (caar gStack))
		   (define/public (mergeTags)
		     (cadar gStack))
		   (define/public (selectionPredicate f)
		     (attr 'selectionPredicate f))
		   (define/public (getSelectionPredicate)
		     (getAttr 'selectionPredicate))
		   (define/public (verboseGrouping p)
		     (attr 'verboseGrouping p)
		     )
		   (define/public (getVerboseGrouping)
		     (getAttr 'verboseGrouping)
		     )
		   
		   (define/public (pushTransform t)
		     (set! tStack (cons (list t (tell1 t thenXformWith (total))) tStack)))
		   
		   (define/public (pushMaterial mat)
		     (let* ([mm (mergeMaterials)])
		       (set! mStack (cons (list mat (if mat 
							(tell1 mat mergeWith mm)
							mm)) mStack))))
		   (define/public (pushTags tl)
		     (set! gStack (cons (list tl (append tl (mergeTags))) gStack)))
		   
		   (define/public (popTags)
		     (if (not (null? (cdr gStack))) (set! gStack (cdr gStack))))
		   
		   (define/public (popTransform)
		     (if (not (null? (cdr tStack))) (set! tStack (cdr tStack))))
		   
		   (define/public (popMaterial)
		     (if (not (null? (cdr mStack))) (set! mStack (cdr mStack))))
		   
		   (define/public (mapPoint p)
		     (tell1 (total) mapPoint p))
		   
		   (define/public (mapDirection v)
		     (tell1 (total) mapDirection v))
		   
		   (define/override (describe)
		     (void)      )
		   
		   (define/public (POVDescribe context)
		     (void))
		   
		   (begin
		     (super-new)
		     (let* ([t (new Transform)]
			    [m defaultMaterial])
		       (set! tStack (list (list t t)))
		       (set! mStack (list (list m m)))
		       (set! gStack (list (list '() '())))
		       (set! d 0)
		       (verboseGrouping #f)
		       )
		     )
		   )
	    )

	  (define Reference
	    (class Primitive
		   (public object getObject)
		   (inherit attr getAttr description getXform getMaterial getTags)
		   
		   (define (object p)
		     (attr 'part p))
		   (define (getObject)
		     (getAttr 'part))
		   
		   (define/override (POVDescribe  context)
		     (tell context 
			   (pushTransform (getXform))
			   (pushMaterial (getMaterial)) (pushTags (getTags)))
		     (tell1 (getObject) POVDescribe context)
		     (tell context
			   (popTransform)
			   (popMaterial)
			   (popTags)))
		   
		   (begin
		     (super-new)
		     (description "A Mead object reference.")
		     )
		   )
	    )

	  (define Group
	    (class Primitive 
		   (inherit description attr getAttr getXform material getMaterial getTags)
					; private:
					; a list of references that make up this part
		   (define/public (forward p msgs)
		     (for-each (lambda (x)
				 (cond ([is-a? x Transform] (tell (tell1 p getXform) (suffixWith x)))
				       ([is-a? x Material] (tell p (material x)))
				       ([symbol? x] (tell p (tag x)))
				       (else (raise-type-error 'add "object, material, or transform" x))
				       )) msgs))
		   
		   (define/public (add o . msgs)
		     (let ((r (new Reference)))
		       (tell1 r object o)
		       (forward r msgs)
		       (attr 'references (append (getAttr 'references) (list (list 'add r))
						 ))
		       r))
		   
		   (define/override (POVDescribe context)
		     (let* ([refs (getAttr 'references)]
			    [t (getXform)]
			    [m (getMaterial)]
			    [g (getTags)]
			    [vg (tell context (getVerboseGrouping))])
		       (begin
			 (tell context (pushMaterial m) (pushTransform t) (pushTags g))
			 (if vg (printf "union {\n"))
			 (for-each (lambda (o) (tell1 (cadr o) POVDescribe context)) refs)
			 (tell context (popTransform) (popMaterial) (popTags))
			 (if vg (printf "}\n"))
			 )))
		   
					; object construction:
		   (begin
		     (super-new)
		     (description "A Mead group object.")
		     (attr 'references null)
		     )
		   
		   )
	    )

;; GLOBAL ROUTINES

	  (define (describe object)
	    (tell1 object describe))

	  (define (translate x y z)
	    (let ([t (new Transform)]) (tell t (translate x y z)) t))
	  (define (xRot a)
	    (let ([t (new Transform)]) (tell t (xRot a)) t))
	  (define (yRot a)
	    (let ([t (new Transform)]) (tell t (yRot a)) t))
	  (define (zRot a)
	    (let ([t (new Transform)]) (tell t (zRot a)) t))
	  (define (scale x y z)
	    (let ([t (new Transform)]) (tell t (scale x y z )) t))
	  (define (xyMirror)
	    (let ([t (new Transform)]) (tell t (xyMirror)) t))
	  (define (xzMirror)
	    (let ([t (new Transform)]) (tell t (xzMirror)) t))
	  (define (yzMirror)
	    (let ([t (new Transform)]) (tell t (yzMirror)) t))
	  (define (compose . xs)
	    (let ([t (new Transform)])
	      (for-each (lambda (x)
					; (printf "composing ") (tell1 t describe) (printf "with\n") (tell1 x describe) (printf "\n")
			  (tell1 t suffixWith x)) xs)
	      t))

	  (define-syntax xform
	    (syntax-rules ()
	      ([xform t ...]
	       (let ([tt (new Transform)])
		 (tell tt t)
		 ...
		 tt))))
	  (define (withoutTag t) 
	    (lambda (x) (not (tell1 x hasTag? t))))

	  (define (withTag t) 
	    (lambda (x) (tell1 x hasTag? t)))
	  
	  (define (clamp x)
	    (- x (floor x)))
	  
	  (define (hsv2rgb c)
	    (let* ([h (* 6 (clamp (/ (car c) 360.0)))]
		   [s (* 1.0 (cadr c))]
		   [v (* 1.0 (caddr c))]
		   [i (floor h)]
		   [f (- h i)]
		   [p (* v (- 1.0 s))]
		   [q (* v (- 1.0 (* s f)))]
		   [t (* v (- 1.0 (* s (- 1 f))))])
	      (cond
	       ([or (= i 6.0) (= i 0.0)] (list v t p))
	       ([= i 1.0] (list q v p))
	       ([= i 2.0] (list p v t))
	       ([= i 3.0] (list p q v))
	       ([= i 4.0] (list t p v))
	       ([= i 5.0] (list v p q))
	       (else i)
	       )
	      )
	    )

	  (define (rgb2hsv c)
	    (let* ([r (car c)]
		   [g (cadr c)]
		   [b (caddr c)]
		   [v (max r g b)]
		   [mn (min r g b)]
		   [range (- v mn)]
		   [h (cond ([= v mn] 0)
			    ([= v r] (+ (* 60 (/ (- g b) range)) (if (>= g b) 0 360)))
			    ([= v g] (+ (* 60 (/ (- b r) range)) 120))
			    (else (+ (* 60 (/ (- r g) range)) 240)))]
		   [s (cond ([= v 0] 0)
			    (else (- 1 (/ mn v))))]
		   )
	      (list h s v)))

	  ;; From Material
	  (define MeadColor
	    (class MeadObject
		   (inherit attr getAttr)
		   (public red green blue alpha filter color
			   getRed getGreen getBlue getAlpha getColor
			   RGB RGBA RGBAF
			   getRGB getRGBA getRGBAF
			   getFilter getTransmit
			   )
		   (define redV 1)
		   (define greenV 1)
		   (define blueV 1)
		   (define alphaV 1)
		   (define filterV 0)
		   (define (color l)
		     (let ((n (length l)))
		       (cond ([<= n 3] (RGB l))
			     ([= n 4] (RGBA l))
			     ([= n 5] (RGBAF l)))))
		   (define (RGB l)
		     (red (car l))
		     (green (cadr l))
		     (blue (caddr l)))
		   (define (RGBA l)
		     (RGB l)
		     (alpha (cadddr l))
		     )
		   (define (RGBAF l)
		     (RGBA l)
		     (filter (cadr (cdddr l))))
		   (define (red r) (set! redV r))
		   (define (blue b) (set! blueV b))
		   (define (green g) (set! greenV g))
		   (define (alpha a) (set! alphaV a))
		   (define (filter t) (set! filterV t))
		   (define (getRed) redV)
		   (define (getBlue) blueV)
		   (define (getGreen) greenV)
		   (define (getAlpha) alphaV)
		   (define (getFilter) filterV)
		   (define (getRGB) (list redV greenV blueV))
		   (define (getRGBA) (list redV greenV blueV alphaV))
		   (define (getRGBAF) (list redV greenV blueV alphaV filterV))
		   (define (getColor) (getRGBAF))
		   (define (getTransmit) (- 1.0 (getAlpha)))
		   (define/public (POVDescribe contextIsIgnored)
		     (printf " rgbft <~a,~a,~a,~a,~a>"
			     (getRed) (getGreen) (getBlue) (getTransmit) (getFilter)))
		   (begin
		     (super-new)
		     (RGBAF '(1 1 1 1 0))
		     )
		   ))
	  
	  (define Material
	    (class MeadObject
		   (inherit attr getAttr)
		   (define/public (pattern t . l)
		     (attr 'pattern t)
		     (let ([n (length l)])
		       (if (> n 0)
			   (if (= n 1)
			       (color (car l))
			       (colors l))))
		     )
		   
		   (define/public (brickPattern brickColor mortarColor)
		     (pattern 'brick brickColor mortarColor)
		     )
		   (define/public (checkerPattern color0 color1)
		     (pattern 'checker color0 color1)
		     )
		   (define/public (hexagonPattern first second third)
		     (pattern 'hexagon first second third)
		     )
		   (define/public (radialPattern r first second . others)
		     (tell/apply this pattern (append (list 'radial first second) others))
		     (frequency r)
		     )
		   (define/public (frequency f)
		     (attr 'frequency f)
		     )
		   (define/public (getFrequency)
		     (getAttr 'frequency)
		     )
		   (define/public (gradientPattern first second . others)
		     (tell/apply this pattern (append (list 'gradient first second) others))
		     )
		   (define/public (granitePattern first second . others)
		     (tell/apply this pattern (append (list 'granite first second) others))
		     )
		   (define/public (cracklePattern first second . others)
		     (tell/apply this pattern (append (list 'crackle first second) others))
		     )
		   (define/public (marblePattern first second . others)
		     (if (not (getTurbulence)) (turbulence .6))
		     (tell/apply this pattern (append (list 'marble first second) others))
		     )
		   (define/public (solid c)
		     (color c)
		     )
		   (define/public (grainImage i)
		     (image i)
		     (pattern 'grain))
		   (define/public (grainPattern i first second . colors)
		     (image i)
		     (tell/apply this pattern (append (list 'grainPattern first second) colors))
		     )
		   (define/public (surfaceImage i)
		     (image i)
		     (pattern 'surface)
		     (UVMapped #t)
		     )
		   (define/public (type t)
		     (case t
		       ([plastic] (ambient 0.1) (diffuse 0.5) (roughness 0.05) (specularity 0.7))
		       ([plaster] (ambient 0.2) (diffuse 0.9) (roughness 1) (specularity 0))
		       ([mirror] (ambient 0) (diffuse 0.1) (reflection 0.9))

		       ([glass] (ambient 0) (diffuse 0.1) (transparency 0.8) (refraction 1.5))
		       )
		     )
		   (define/public (getPattern)
		     (getAttr 'pattern)
		     )
		   (define/public (UVMapped bool)
		     (attr 'UVMapped bool)
		     )
		   (define/public (getUVMapped)
		     (getAttr 'UVMapped)
		     )
		   ;; Items having to do with POV "pigment":
		   (define/public (color c)
		     (if (list? c)
			 (let ([c2 (new MeadColor)])
			   (tell1 c2 color c)
			   (attr 'color c2))
			 (attr 'color c)
			 )
		     )
		   (define/public (getColor)
		     (getAttr 'color)
		     )
		   (define/public (image i)
		     (attr 'image i)
		     )
		   (define/public (getImage)
		     (getAttr 'image)
		     )
		   (define/public (turbulence t)
		     (attr 'turbulence t)
		     )
		   (define/public (getTurbulence)
		     (getAttr 'turbulence)
		     )
		   (define/public (transparency t)
		     (attr 'transparency t)
		     )
		   (define/public (getTransparency)
		     (getAttr 'transparency)
		     )
		   (define/public (hasXform?)
		     (getAttr 'xForm)
		     )
		   (define/public (getXform)
		     (if (not (hasXform?)) (attr 'xForm (new Transform)))
		     (getAttr 'xForm)
		     )
		   (define/public (xRot angle) (tell1 (getXform) xRot angle))
		   (define/public (yRot angle) (tell1 (getXform) yRot angle))
		   (define/public (zRot angle) (tell1 (getXform) zRot angle))
		   (define/public (scale x y z) (tell1 (getXform) scale x y z))
		   (define/public (translate x y z) (tell1 (getXform) translate x y z))
		   (define/public (xyMirror) (tell1 (getXform) xyMirror))
		   (define/public (xzMirror) (tell1 (getXform) xzMirror))
		   (define/public (yzMirror) (tell1 (getXform) yzMirror))
		   
		   (define/public (imageType t)
		     (attr 'imageType t)
		     )
		   (define/public (getImageType)
		     (getAttr 'imageType))
		   (define/public (colors l)
		     (attr 'colors (map (lambda (c) (if (list? c) (let ([d (new MeadColor)])
								    (tell1 d color c)
								    d) c)) l)))
		   (define/public (getColors)
		     (getAttr 'colors))
		   ;; Items having to do with POV "finish":
		   (define/public (phong e)
		     (attr 'phong e))
		   (define/public (getPhong)
		     (getAttr 'phong))
		   (define/public (ambient coeff)
		     (attr 'ambient coeff))
		   (define/public (getAmbient)
		     (getAttr 'ambient))
		   (define/public (diffuse coeff)
		     (attr 'diffuse coeff))
		   (define/public (getDiffuse)
		     (getAttr 'diffuse))
		   
		   (define/public (specularity coeff)
		     (attr 'specularity coeff)
		     )
		   (define/public (getSpecularity)
		     (getAttr 'specularity))
		   (define/public (roughness coeff)
		     (attr 'roughness coeff))
		   (define/public (getRoughness)
		     (getAttr 'roughness))
		   ;; Items having to do with POV "Interior"
		   (define/public (refraction index)
		     (attr 'refraction index)
		     )
		   (define/public (getRefraction)
		     (getAttr 'refraction)
		     )
		   (define/public (caustics index)
		     (attr 'caustics index)
		     )
		   (define/public (getCaustics)
		     (getAttr 'getCaustics))
		   (define/public (fadeDistance d)
		     (attr 'fadeDistance d))
		   (define/public (getFadeDistance)
		     (getAttr 'fadeDistance))
		   (define/public (fadePower p)
		     (attr 'fadePower p))
		   (define/public (getFadePower)
		     (getAttr 'fadePower))

		   (define/public (reflection coeff)
		     (attr 'reflection coeff)
		     )
		   (define/public (getReflection)
		     (getAttr 'reflection))
		   (define/public (metallic bool)
		     (attr 'metallic bool))
		   (define/public (getMetallic)
		     (getAttr 'metallic))
		   (define/public (mergeWith m)
		     (let* ([nm (new Material)])
		       (tell1 nm attributes (append (tell1 this getAttributes) (tell1 m getAttributes)))
		       nm))
		   (define/public (POVDescribeTurbulence t)
		     (printf " warp{turbulence ~a}" t)
		     )
		   (define/public (POVDescribeColors l)
		     (map (lambda (x) (tell1 x POVDescribe #f)) l))
		   (define/public (POVDescribeCM l)
		     (printf "color_map {")
		     (let ([n (- (length l) 1)])
		       (do ([i 0.0 (+ 1.0 i)]
			    [l l (cdr l)])
			   ((null? l))
			 (printf " [~a " (/ i n))
			 (tell1 (car l) POVDescribe #f)
			 (printf "]")
			 )
		       )
		     (printf "}\n")
		     )
		   (define/public (makeImagePath f)
		     (let* ([home (or (getenv "HOME") "/tmp")]
			    [ef (tell environment (getImageFolder))]
			    [folder (or (and (string? ef) ef) "Images")]
			    [file (or (and (string? f) (< 0 (string-length f)) f) "untitled.png")])
		       (if (and (< 0 (string-length file)) (eq? #\/ (string-ref file 0)))
			   (begin
			     (set! home "")
			     (set! folder "")))
		       (if (and (< 0 (string-length folder)) (eq? #\/ (string-ref folder 0)))
			   (begin
			     (set! home "")))
		       (string-append home "/" folder "/" file)
		       )
		     )


		   (define/public (POVDescribePigment)
		     (printf "pigment{")
		     (if (getUVMapped) (printf "uv_mapping "))
		     (case (getPattern)
		       ((grain surface)
			(printf "image_map{~a \"~a\"" (getImageType) (tell environment (makeImagePath (getImage))))
			(if (getTransparency) (printf " transmit all ~a" (getTransparency)))
			(printf "}"))
		       ((radial)
			(printf "radial frequency ~a " (getFrequency))
			(POVDescribeCM (getColors)))
		       ((grainPattern)
			(printf "image_pattern{~a \"~a\"" (getImageType) (tell environment (makeImagePath (getImage))))
         ;;; FIXME (if (getOneShot) (printf " once"))
			(printf "}")
			(POVDescribeCM (getColors)))
		       ((checker brick hexagon)
			(printf "~a " (getPattern))
			(POVDescribeColors (getColors)))
		       ((gradient)
			(printf "gradient x ")
			(POVDescribeCM (getColors)))
		       ((granite crackle marble)
			(printf "~a " (getPattern))
			(POVDescribeCM (getColors)))
		       (else
			(printf "color ")
			(tell1 (getColor) POVDescribe #f)
			(if (and (getTransparency) (= 0 (tell1 (getColor) getTransmit)))
			    (printf " transmit ~a" (getTransparency)))
			)
		       )
		     (if (getTurbulence) (POVDescribeTurbulence (getTurbulence)))
		     (printf "}")
		     )
		   
		   (define/public (POVDescribeFinish)
		     (let* ([a (getAmbient)]
			    [p (getPhong)]
			    [df (getDiffuse)]
			    [sp (getSpecularity)]
			    [r (getRoughness)]
			    [rf (getReflection)]
			    [mt (getMetallic)]
			    )

		       (if (or a df sp r rf mt p)
			   (begin
			     (printf "finish { ")
			     (if a (printf " ambient ~a" a))
			     (if df (printf " diffuse ~a" df))
			     (if sp (printf " specular ~a" sp))
			     (if r (printf " roughness ~a" r))
			     (if (or rf mt)
				 (begin
				   (printf " reflection { ")
				   (if rf (printf "~a" rf))
				   (if mt (printf " metallic"))
				   (printf "}")))
			     (if p (printf " phong ~a" p))
			     (printf "} ")))
		       ))
		   
		   (define/public (POVDescribeInterior)
		     (let* ([ior (getRefraction)]
			    [cau (getCaustics)]
			    [fd (getFadeDistance)]
			    [fp (getFadePower)]
			    )
		       (if (or ior cau fd fp)
			   (begin
			     (printf "interior { ")
			     (if ior (printf " ior ~a" ior))
			     (if cau (printf " caustics ~a" cau))
			     (if fd (printf " fade_distance ~a" fd))
			     (if fp (printf " fade_power ~a" fp))
			     (printf "}")
			     )
			   )
		       )
		     )
		   
		   (define/public (POVDescribe . contextIsIgnored)
		     (printf "texture{")
		     (POVDescribePigment)
		     (POVDescribeFinish)
		     (if (hasXform?)
			 (let ([x (getXform)])
			   (tell1 x POVDescribe #f))
			 )
		     (printf "}\n")
		     (POVDescribeInterior)
		     )
		   
		   (begin
		     (super-new)
		     (imageType 'png)
					;(pattern 'solid)
		     (attr 'xForm (new Transform))
		     (scale 100 100 100)
		     (translate -50 -50 -50)
		     )
		   ;; (phong 1)
		   )
	    )

	  (define emptyMaterial (new Material))
	  (object defaultMaterial Material (color '(.5 .5 .5)))
	  )
