(module Surfaces mzscheme
  (require (lib "Syntax.ss" "Mead")
           (lib "Math.ss" "Mead")
           (lib "Object.ss" "Mead")) ;; this incorporates Materials, as well.
  (provide Surface QuadricSurface CubicSurface QuarticSurface
           InfiniteCylinder InfiniteCone InfiniteParaboloid InfiniteHyperboloid InfiniteSaddle
           Lemniscate Piriform TorusSurface)
  
(define coeffIndicies
  '#( null
      null
      ((x2 0) (xy 1) (xz 2) (x 3) (y2 4) (yz 5) (y 6) (z2 7) (z 8) (1 9)
              (A 0) (B 4) (C 7) (D 1) (E 2) (F 5) (G 3) (H 6) (I 8) (J 9))
      ((x3 0) (x2y 1) (x2z 2) (x2 3) (xy2 4) (xyz 5) (xy 6) (xz2 7) (xz 8) (x 9) (y3 10) (y2z 11) (y2 12) (yz2 13) (yz 14) (y 15) (z3 16) (z2 17) (z 18) (1 19))
      ((x4 0) (x3y 1) (x3z 2) (x3 3) (x2y2 4) (x2yz 5) (x2y 6) (x2z2 7) (x2z 8) (x2 9) (xy3 10) (xy2z 11) (xy2 12) (xyz2 13) (xyz 14) (xy 15) (xz3 16) (xz2 17) (xz 18) (x 19) (y4 20) (y3z 21) (y3 22) (y2z2 23) (y2z 24) (y2 25) (yz3 26) (yz2 27) (yz 28) (y 29) (z4 30) (z3 31) (z2 32) (z 33) (1 34))))

(define Surface
  (class Primitive
    (inherit selected attr getAttr material getMaterial getXform description)
    (define/public (degree n)
      (let ([nCoeffs (/ (* (+ n 1) (+ n 2) (+ n 3)) 6)])
        (attr 'degree n)
        (attr 'nCoeffs nCoeffs)
	(attr 'coeffIndicies (vector-ref coeffIndicies n))
        (attr 'coefficients (make-vector nCoeffs 0))))
    (define/public (getDegree)
      (getAttr 'degree))
    (define/public (getCoefficients)
      (getAttr 'coefficients))
    (define/public (coefficients p)
      (let ([c (getCoefficients)]
            [ci (getAttr 'coeffIndicies)])
	(let loop ([p p])
	  (if (not (null? p))
	      (let* ([term (car p)]
		     [coeff (car term)]
		     [power (cadr term)]
		     [idx (cadr (assoc power ci))])
		(vector-set! c idx coeff)
		(loop (cdr p)))))))
    
    (define/public (POVPrimitiveDescribe context)
      (super POVDescribe context)
      ) 
    
    (define/override (POVDescribe context)
      (let ([d (getDegree)]
            [nc (getAttr 'nCoeffs)]
            [c (getCoefficients)]
            )
        (printf "poly {~a,<" d)
        (do ([i 0 (+ i 1)])
          ((>= i nc))
          (if (> i 0) (printf ","))
          (printf "~a" (vector-ref c i)))
        (printf ">\n")
        (super POVDescribe context)
        (printf "}\n")))
    
    (begin
      (super-new)
      (description "A primitive surface.")
      )
    )
  )

(define QuadricSurface
  (class Surface
    (inherit degree description POVPrimitiveDescribe getCoefficients)
    
    (define/override (POVDescribe context)
      (let* ([v (getCoefficients)]
             [degree 2]
             [ci (vector-ref coeffIndicies degree)]
             [A (vector-ref v (cadr (assoc 'A ci)))]
             [B (vector-ref v (cadr (assoc 'B ci)))]
             [C (vector-ref v (cadr (assoc 'C ci)))]
             [D (vector-ref v (cadr (assoc 'D ci)))]
             [E (vector-ref v (cadr (assoc 'E ci)))]
             [F (vector-ref v (cadr (assoc 'F ci)))]
             [G (vector-ref v (cadr (assoc 'G ci)))]
             [H (vector-ref v (cadr (assoc 'H ci)))]
             [I (vector-ref v (cadr (assoc 'I ci)))]
             [J (vector-ref v (cadr (assoc 'J ci)))])
        
        (printf "quadric {")
        (printf "<~a,~a,~a>,<~a,~a,~a>,<~a,~a,~a>,~a\n" A B C D E F G H I J)
        (POVPrimitiveDescribe context)
        (printf "}\n")
        ))
    
    (begin
      (super-new)
      (degree 2)
      (description "A quadric surface.")
      )
    )
  )

(define CubicSurface
  (class Surface
    (inherit degree description)
    
    (begin
      (super-new)
      (degree 3)
      (description "A cubic surface."))))

(define QuarticSurface
  (class Surface
    (inherit degree description)
    
    (begin
      (super-new)
      (degree 4)
      (description "A quartic surface."))))

(define InfiniteCylinder
  (class QuadricSurface
    (inherit description coefficients)
    (begin
      (super-new)
      (coefficients '((1 x2) (1 z2) (-10000 1)))
      (description "An infinite cylinder.")
      )
    )
  )

(define InfiniteCone
  (class QuadricSurface
    (inherit description coefficients)
    (begin
      (super-new)
      (coefficients '((1 x2) (-1 y2) (1 z2)))
      (description "An infinite cone.")
      )
    )
  )

(define InfiniteParaboloid
  (class QuadricSurface
    (inherit description coefficients)
    (begin
      (super-new)
      (coefficients '((1 x2) (-100 y) (1 z2)))
      (description "An infinite paraboloid.")
      )
    )
  )

(define InfiniteHyperboloid
  (class QuadricSurface
    (inherit description coefficients)
    (begin
      (super-new)
      (coefficients '((1 x2) (-1 y2) (1 z2) (-1 1)))
      (description "An infinite hyperboloid.")
      )
    )
  )

(define InfiniteSaddle
  (class QuarticSurface
    (inherit description coefficients)
    (begin
      (super-new)
      (coefficients '((1 x3) (-1 y3) (-1 z)))
      (description "A saddle curve.")
      )
    )
  )

(define Lemniscate
  (class QuarticSurface
    (inherit description coefficients)
    (begin
      (super-new)
      (coefficients '((1 y4) (-1 y2) (1 x2) (1 z2)))
      (description "A lemniscate.")
      )
    )
  )

(define Piriform
  (class QuarticSurface
    (inherit description coefficients)
    (begin
      (super-new)
      (coefficients '((1 x4) (-1 x3) (1 x2) (1 z2)))
      (description "A piriform.")
      )
    )
  )

(define TorusSurface
  (class QuarticSurface
    (inherit attr getAttr degree coefficients description)
    (define/public (radii inner outer)
      (let* ([so (sqr outer)]
             [si (sqr inner)]
             [sum (+ so si)]
             [diff (- so si)])
        (attr 'radii (list inner outer))
        (coefficients (list '(1 x4) '(1 y4) '(1 z4) '(2 x2y2) '(2 x2z2) '(2 y2z2)
                            (list (* -2 sum) 'x2)
                            (list (* 2  diff) 'y2)
                            (list (* -2 sum) 'z2)
                            (list (sqr diff) 1)))))
    
    (define/public (getRadii)
      (getAttr 'radii))
    
    (begin
      (super-new)
      (radii (sqrt 40) (sqrt 12))
      (description "A toroidal surface.")
      )
    )
  )
)
